@@ -22,14 +22,17 @@ static int list_count(struct list_head *h)
 	return c;
 }
 
-static int stats_main(void *ctx, struct apk_database *db, struct apk_string_array *args)
+static int stats_main(void *ctx, struct apk_ctx *ac, struct apk_string_array *args)
 {
-	printf(
+	struct apk_out *out = &ac->out;
+	struct apk_database *db = ac->db;
+
+	apk_out(out,
 		"installed:\n"
 		"  packages: %d\n"
 		"  dirs: %d\n"
 		"  files: %d\n"
-		"  bytes: %zu\n"
+		"  bytes: %" PRIu64 "\n"
 		"  triggers: %d\n"
 		"available:\n"
 		"  names: %d\n"
