@@ -33,6 +33,8 @@ struct apk_solver_state {
 	struct apk_changeset *changeset;
 	struct list_head dirty_head;
 	struct list_head unresolved_head;
+	struct list_head selectable_head;
+	struct list_head resolvenow_head;
 	unsigned int errors;
 	unsigned int solver_flags_inherit;
 	unsigned int pinning_inherit;
@@ -50,10 +52,8 @@ void apk_solver_set_name_flags(struct apk_name *name,
 			       unsigned short solver_flags,
 			       unsigned short solver_flags_inheritable)
 {
-	struct apk_provider *p;
-
 	name->solver_flags_set = 1;
-	foreach_array_item(p, name->providers) {
+	apk_array_foreach(p, name->providers) {
 		struct apk_package *pkg = p->pkg;
 		dbg_printf("marking '" PKG_VER_FMT "' = 0x%04x / 0x%04x\n",
 			PKG_VER_PRINTF(pkg), solver_flags, solver_flags_inheritable);
@@ -99,27 +99,51 @@ static void queue_dirty(struct apk_solver_state *ss, struct apk_name *name)
 	list_add_tail(&name->ss.dirty_list, &ss->dirty_head);
 }
 
-static void queue_unresolved(struct apk_solver_state *ss, struct apk_name *name)
+static bool queue_resolvenow(struct apk_name *name)
 {
-	int want;
+	return name->ss.reverse_deps_done && name->ss.requirers &&
+		name->ss.has_auto_selectable && !name->ss.has_options;
+}
 
-	if (name->ss.locked)
+static void queue_insert(struct list_head *head, struct apk_name *name)
+{
+	struct apk_name *name0;
+
+	list_for_each_entry(name0, head, ss.unresolved_list) {
+		if (name->ss.order_id < name0->ss.order_id) continue;
+		list_add_before(&name->ss.unresolved_list, &name0->ss.unresolved_list);
 		return;
+	}
+	list_add_tail(&name->ss.unresolved_list, head);
+}
 
-	want = (name->ss.requirers > 0) || (name->ss.has_iif);
-	dbg_printf("queue_unresolved: %s, want=%d (requirers=%d, has_iif=%d)\n", name->name, want, name->ss.requirers, name->ss.has_iif);
-	if (want && !list_hashed(&name->ss.unresolved_list))
-		list_add(&name->ss.unresolved_list, &ss->unresolved_head);
-	else if (!want && list_hashed(&name->ss.unresolved_list))
+static void queue_unresolved(struct apk_solver_state *ss, struct apk_name *name, bool reevaluate)
+{
+	if (name->ss.locked) return;
+	if (list_hashed(&name->ss.unresolved_list)) {
+		if (name->ss.resolvenow) return;
+		if (queue_resolvenow(name) == 1)
+			name->ss.resolvenow = 1;
+		else if (!reevaluate)
+			return;
 		list_del_init(&name->ss.unresolved_list);
+	} else {
+		if (name->ss.requirers == 0 && !name->ss.has_iif && !name->ss.iif_needed) return;
+		name->ss.resolvenow = queue_resolvenow(name);
+	}
+
+	dbg_printf("queue_unresolved: %s, requirers=%d, has_iif=%d, resolvenow=%d\n",
+		name->name, name->ss.requirers, name->ss.has_iif, name->ss.resolvenow);
+	if (name->ss.resolvenow) {
+		list_add_tail(&name->ss.unresolved_list, &ss->resolvenow_head);
+		return;
+	}
+	queue_insert(name->ss.has_auto_selectable ? &ss->selectable_head : &ss->unresolved_head, name);
 }
 
 static void reevaluate_reverse_deps(struct apk_solver_state *ss, struct apk_name *name)
 {
-	struct apk_name **pname0, *name0;
-
-	foreach_array_item(pname0, name->rdepends) {
-		name0 = *pname0;
+	apk_array_foreach_item(name0, name->rdepends) {
 		if (!name0->ss.seen) continue;
 		name0->ss.reevaluate_deps = 1;
 		queue_dirty(ss, name0);
@@ -128,10 +152,7 @@ static void reevaluate_reverse_deps(struct apk_solver_state *ss, struct apk_name
 
 static void reevaluate_reverse_installif(struct apk_solver_state *ss, struct apk_name *name)
 {
-	struct apk_name **pname0, *name0;
-
-	foreach_array_item(pname0, name->rinstall_if) {
-		name0 = *pname0;
+	apk_array_foreach_item(name0, name->rinstall_if) {
 		if (!name0->ss.seen) continue;
 		if (name0->ss.no_iif) continue;
 		name0->ss.reevaluate_iif = 1;
@@ -141,79 +162,67 @@ static void reevaluate_reverse_installif(struct apk_solver_state *ss, struct apk
 
 static void reevaluate_reverse_installif_pkg(struct apk_solver_state *ss, struct apk_package *pkg)
 {
-	struct apk_dependency *d;
 	reevaluate_reverse_installif(ss, pkg->name);
-	foreach_array_item(d, pkg->provides)
+	apk_array_foreach(d, pkg->provides)
 		reevaluate_reverse_installif(ss, d->name);
 }
 
 static void disqualify_package(struct apk_solver_state *ss, struct apk_package *pkg, const char *reason)
 {
-	struct apk_dependency *p;
-
 	dbg_printf("disqualify_package: " PKG_VER_FMT " (%s)\n", PKG_VER_PRINTF(pkg), reason);
 	pkg->ss.pkg_selectable = 0;
 	reevaluate_reverse_deps(ss, pkg->name);
-	foreach_array_item(p, pkg->provides)
+	apk_array_foreach(p, pkg->provides)
 		reevaluate_reverse_deps(ss, p->name);
 	reevaluate_reverse_installif_pkg(ss, pkg);
 }
 
-static int dependency_satisfiable(struct apk_solver_state *ss, struct apk_dependency *dep)
+static bool dependency_satisfiable(struct apk_solver_state *ss, const struct apk_package *dpkg, struct apk_dependency *dep)
 {
 	struct apk_name *name = dep->name;
-	struct apk_provider *p;
 
-	if (dep->conflict && ss->ignore_conflict)
-		return TRUE;
+	if (apk_dep_conflict(dep) && ss->ignore_conflict) return true;
+	if (name->ss.locked) return apk_dep_is_provided(dpkg, dep, &name->ss.chosen);
+	if (name->ss.requirers == 0 && apk_dep_is_provided(dpkg, dep, &provider_none))
+		return true;
 
-	if (name->ss.locked)
-		return apk_dep_is_provided(dep, &name->ss.chosen);
+	apk_array_foreach(p, name->providers)
+		if (p->pkg->ss.pkg_selectable && apk_dep_is_provided(dpkg, dep, p))
+			return true;
 
-	if (name->ss.requirers == 0 && apk_dep_is_provided(dep, &provider_none))
-		return TRUE;
-
-	foreach_array_item(p, name->providers)
-		if (p->pkg->ss.pkg_selectable && apk_dep_is_provided(dep, p))
-			return TRUE;
-
-	return FALSE;
+	return false;
 }
 
 static void discover_name(struct apk_solver_state *ss, struct apk_name *name)
 {
 	struct apk_database *db = ss->db;
-	struct apk_name **pname0;
-	struct apk_provider *p;
-	struct apk_dependency *dep;
 	unsigned int repos, num_virtual = 0;
 
-	if (name->ss.seen)
-		return;
+	if (name->ss.seen) return;
 
 	name->ss.seen = 1;
 	name->ss.no_iif = 1;
-	foreach_array_item(p, name->providers) {
+	apk_array_foreach(p, name->providers) {
 		struct apk_package *pkg = p->pkg;
 		if (!pkg->ss.seen) {
 			pkg->ss.seen = 1;
 			pkg->ss.pinning_allowed = APK_DEFAULT_PINNING_MASK;
 			pkg->ss.pinning_preferred = APK_DEFAULT_PINNING_MASK;
-			pkg->ss.pkg_available = pkg->filename_ndx ||
-				(pkg->repos & db->available_repos & ~BIT(APK_REPOSITORY_CACHED));
+			pkg->ss.pkg_available = pkg->filename_ndx || apk_db_pkg_available(db, pkg);
 			/* Package is in 'cached' repository if filename is provided,
 			 * or it's a 'virtual' package with install_size zero */
-			pkg->ss.pkg_selectable =
-				(pkg->repos & db->available_repos) ||
-				pkg->cached_non_repository ||
-				pkg->installed_size == 0 ||
-				pkg->ipkg;
+			pkg->ss.pkg_selectable = !pkg->uninstallable &&
+				(BIT(pkg->layer) & db->active_layers) &&
+				(pkg->ss.pkg_available ||
+				 pkg->cached || pkg->filename_ndx ||
+				 pkg->cached_non_repository ||
+				 pkg->installed_size == 0 ||  pkg->ipkg);
 
 			/* Prune install_if packages that are no longer available,
 			 * currently works only if SOLVERF_AVAILABLE is set in the
 			 * global solver flags. */
 			pkg->ss.iif_failed =
-				(pkg->install_if->num == 0) ||
+				(apk_array_len(pkg->install_if) == 0) ||
 				((ss->solver_flags_inherit & APK_SOLVERF_AVAILABLE) &&
 				 !pkg->ss.pkg_available);
 
@@ -226,7 +235,7 @@ static void discover_name(struct apk_solver_state *ss, struct apk_name *name)
 				pkg->cached_non_repository ||
 				pkg->ipkg;
 
-			foreach_array_item(dep, pkg->depends)
+			apk_array_foreach(dep, pkg->depends)
 				discover_name(ss, dep->name);
 
 			dbg_printf("discover " PKG_VER_FMT ": tag_ok=%d, tag_pref=%d selectable=%d\n",
@@ -240,27 +249,34 @@ static void discover_name(struct apk_solver_state *ss, struct apk_name *name)
 		num_virtual += (p->pkg->name != name);
 	}
 
-	foreach_array_item(p, name->providers) {
+	apk_array_foreach_item(name0, name->rinstall_if)
+		discover_name(ss, name0);
+
+	apk_array_foreach(p, name->providers) {
 		struct apk_package *pkg = p->pkg;
-		foreach_array_item(pname0, pkg->name->rinstall_if)
-			discover_name(ss, *pname0);
-		foreach_array_item(dep, pkg->provides) {
-			if (dep->name->ss.seen) continue;
+		apk_array_foreach_item(name0, pkg->name->rinstall_if)
+			discover_name(ss, name0);
+		apk_array_foreach(dep, pkg->provides)
 			discover_name(ss, dep->name);
-			foreach_array_item(pname0, dep->name->rinstall_if)
-				discover_name(ss, *pname0);
-		}
 	}
 
-	name->ss.order_id = ++ss->order_id;
+	unsigned int order_flags = 0;
+	if (!name->solver_flags_set) order_flags |= 1UL << 31;
+	if (apk_array_len(name->providers) != num_virtual) order_flags |= 1UL << 30;
+	name->ss.order_id = order_flags | ++ss->order_id;
 
-	dbg_printf("discover %s: no_iif=%d num_virtual=%d, order_id=%d\n",
+	apk_array_foreach(p, name->providers) {
+		apk_array_foreach(dep, p->pkg->install_if)
+			discover_name(ss, dep->name);
+	}
+
+	dbg_printf("discover %s: no_iif=%d num_virtual=%d, order_id=%#x\n",
 		name->name, name->ss.no_iif, num_virtual, name->ss.order_id);
 }
 
 static void name_requirers_changed(struct apk_solver_state *ss, struct apk_name *name)
 {
-	queue_unresolved(ss, name);
+	queue_unresolved(ss, name, false);
 	reevaluate_reverse_installif(ss, name);
 	queue_dirty(ss, name);
 }
@@ -293,26 +309,25 @@ static void inherit_pinning_and_flags(
 static void apply_constraint(struct apk_solver_state *ss, struct apk_package *ppkg, struct apk_dependency *dep)
 {
 	struct apk_name *name = dep->name;
-	struct apk_provider *p0;
 	int is_provided;
 
 	dbg_printf("    apply_constraint: %s%s%s" BLOB_FMT "\n",
-		dep->conflict ? "!" : "",
+		apk_dep_conflict(dep) ? "!" : "",
 		name->name,
-		apk_version_op_string(dep->result_mask),
+		apk_version_op_string(dep->op),
 		BLOB_PRINTF(*dep->version));
 
-	if (dep->conflict && ss->ignore_conflict)
+	if (apk_dep_conflict(dep) && ss->ignore_conflict)
 		return;
 
-	name->ss.requirers += !dep->conflict;
-	if (name->ss.requirers == 1 && !dep->conflict)
+	name->ss.requirers += !apk_dep_conflict(dep);
+	if (name->ss.requirers == 1 && !apk_dep_conflict(dep))
 		name_requirers_changed(ss, name);
 
-	foreach_array_item(p0, name->providers) {
+	apk_array_foreach(p0, name->providers) {
 		struct apk_package *pkg0 = p0->pkg;
 
-		is_provided = apk_dep_is_provided(dep, p0);
+		is_provided = apk_dep_is_provided(ppkg, dep, p0);
 		dbg_printf("    apply_constraint: provider: %s-" BLOB_FMT ": %d\n",
 			pkg0->name->name, BLOB_PRINTF(*p0->version), is_provided);
 
@@ -327,19 +342,13 @@ static void apply_constraint(struct apk_solver_state *ss, struct apk_package *pp
 
 static void exclude_non_providers(struct apk_solver_state *ss, struct apk_name *name, struct apk_name *must_provide, int skip_virtuals)
 {
-	struct apk_provider *p;
-	struct apk_dependency *d;
-
-	if (name == must_provide || ss->ignore_conflict)
-		return;
-
+	if (name == must_provide || ss->ignore_conflict) return;
 	dbg_printf("%s must provide %s (skip_virtuals=%d)\n", name->name, must_provide->name, skip_virtuals);
-
-	foreach_array_item(p, name->providers) {
+	apk_array_foreach(p, name->providers) {
 		if (p->pkg->name == must_provide || !p->pkg->ss.pkg_selectable ||
 		    (skip_virtuals && p->version == &apk_atom_null))
 			goto next;
-		foreach_array_item(d, p->pkg->provides)
+		apk_array_foreach(d, p->pkg->provides)
 			if (d->name == must_provide || (skip_virtuals && d->version == &apk_atom_null))
 				goto next;
 		disqualify_package(ss, p->pkg, "provides transitivity");
@@ -364,25 +373,21 @@ static inline int merge_index_complete(unsigned short *index, int num_options)
 	return ret;
 }
 
-static int is_provider_auto_selectable(struct apk_provider *p)
+static bool is_provider_auto_selectable(struct apk_provider *p)
 {
-	/* Virtual packages without provider_priority cannot be autoselected,
-	 * without provider_priority or auto_select_virtual override */
-	if (p->version != &apk_atom_null) return 1;
-	if (p->pkg->provider_priority) return 1;
-	if (p->pkg->name->auto_select_virtual) return 1;
-	if (p->pkg->name->ss.requirers) return 1;
-	return 0;
+	// Virtual packages without provider_priority cannot be autoselected without provider_priority
+	if (p->version != &apk_atom_null) return true;
+	if (p->pkg->provider_priority) return true;
+	if (p->pkg->name->ss.requirers) return true;
+	return false;
 }
 
 static void reconsider_name(struct apk_solver_state *ss, struct apk_name *name)
 {
-	struct apk_name *name0, **pname0;
-	struct apk_dependency *dep;
 	struct apk_package *first_candidate = NULL, *pkg;
-	struct apk_provider *p;
 	int reevaluate_deps, reevaluate_iif;
 	int num_options = 0, num_tag_not_ok = 0, has_iif = 0, no_iif = 1;
+	bool reevaluate = false;
 
 	dbg_printf("reconsider_name: %s\n", name->name);
 
@@ -390,19 +395,19 @@ static void reconsider_name(struct apk_solver_state *ss, struct apk_name *name)
 	reevaluate_iif = name->ss.reevaluate_iif;
 	name->ss.reevaluate_deps = 0;
 	name->ss.reevaluate_iif = 0;
-	name->ss.has_auto_selectable = 0;
 
 	/* propagate down by merging common dependencies and
 	 * applying new constraints */
-	foreach_array_item(p, name->providers) {
+	unsigned int has_auto_selectable = 0;
+	apk_array_foreach(p, name->providers) {
 		/* check if this pkg's dependencies have become unsatisfiable */
 		pkg = p->pkg;
 		pkg->ss.dependencies_merged = 0;
 		if (reevaluate_deps) {
 			if (!pkg->ss.pkg_selectable)
 				continue;
-			foreach_array_item(dep, pkg->depends) {
-				if (!dependency_satisfiable(ss, dep)) {
+			apk_array_foreach(dep, pkg->depends) {
+				if (!dependency_satisfiable(ss, pkg, dep)) {
 					disqualify_package(ss, pkg, "dependency no longer satisfiable");
 					break;
 				}
@@ -416,13 +421,17 @@ static void reconsider_name(struct apk_solver_state *ss, struct apk_name *name)
 		     pkg->ss.iif_failed == 0)) {
 			pkg->ss.iif_triggered = 1;
 			pkg->ss.iif_failed = 0;
-			foreach_array_item(dep, pkg->install_if) {
+			apk_array_foreach(dep, pkg->install_if) {
 				if (!dep->name->ss.locked) {
+					if (apk_dep_conflict(dep)) {
+						dep->name->ss.iif_needed = true;
+						queue_unresolved(ss, dep->name, false);
+					}
 					pkg->ss.iif_triggered = 0;
 					pkg->ss.iif_failed = 0;
 					break;
 				}
-				if (!apk_dep_is_provided(dep, &dep->name->ss.chosen)) {
+				if (!apk_dep_is_provided(pkg, dep, &dep->name->ss.chosen)) {
 					pkg->ss.iif_triggered = 0;
 					pkg->ss.iif_failed = 1;
 					break;
@@ -430,7 +439,7 @@ static void reconsider_name(struct apk_solver_state *ss, struct apk_name *name)
 			}
 		}
 		if (reevaluate_iif && pkg->ss.iif_triggered) {
-			foreach_array_item(dep, pkg->install_if)
+			apk_array_foreach(dep, pkg->install_if)
 				inherit_pinning_and_flags(ss, pkg, dep->name->ss.chosen.pkg);
 		}
 		has_iif |= pkg->ss.iif_triggered;
@@ -438,7 +447,7 @@ static void reconsider_name(struct apk_solver_state *ss, struct apk_name *name)
 		dbg_printf("  "PKG_VER_FMT": iif_triggered=%d iif_failed=%d, no_iif=%d\n",
 			PKG_VER_PRINTF(pkg), pkg->ss.iif_triggered, pkg->ss.iif_failed,
 			no_iif);
-		name->ss.has_auto_selectable |= pkg->ss.iif_triggered;
+		has_auto_selectable |= pkg->ss.iif_triggered;
 
 		if (name->ss.requirers == 0)
 			continue;
@@ -449,41 +458,44 @@ static void reconsider_name(struct apk_solver_state *ss, struct apk_name *name)
 			first_candidate = pkg;
 
 		/* FIXME: can merge also conflicts */
-		foreach_array_item(dep, pkg->depends)
-			if (!dep->conflict)
+		apk_array_foreach(dep, pkg->depends)
+			if (!apk_dep_conflict(dep))
 				merge_index(&dep->name->ss.merge_depends, num_options);
 
 		if (merge_index(&pkg->name->ss.merge_provides, num_options))
 			pkg->name->ss.has_virtual_provides |= (p->version == &apk_atom_null);
-		foreach_array_item(dep, pkg->provides)
+		apk_array_foreach(dep, pkg->provides)
 			if (merge_index(&dep->name->ss.merge_provides, num_options))
 				dep->name->ss.has_virtual_provides |= (dep->version == &apk_atom_null);
 
 		num_tag_not_ok += !pkg->ss.tag_ok;
 		num_options++;
-		if (!name->ss.has_auto_selectable && is_provider_auto_selectable(p))
-			name->ss.has_auto_selectable = 1;
+		if (!has_auto_selectable && is_provider_auto_selectable(p))
+			has_auto_selectable = 1;
 	}
 	name->ss.has_options = (num_options > 1 || num_tag_not_ok > 0);
 	name->ss.has_iif = has_iif;
 	name->ss.no_iif = no_iif;
-	queue_unresolved(ss, name);
+	if (has_auto_selectable != name->ss.has_auto_selectable) {
+		name->ss.has_auto_selectable = has_auto_selectable;
+		reevaluate = true;
+	}
 
 	if (first_candidate != NULL) {
 		pkg = first_candidate;
-		foreach_array_item(p, name->providers)
+		apk_array_foreach(p, name->providers)
 			p->pkg->ss.dependencies_used = p->pkg->ss.dependencies_merged;
 
 		/* propagate down common dependencies */
 		if (num_options == 1) {
 			/* FIXME: keeps increasing counts, use bit fields instead? */
-			foreach_array_item(dep, pkg->depends)
+			apk_array_foreach(dep, pkg->depends)
 				if (merge_index_complete(&dep->name->ss.merge_depends, num_options))
 					apply_constraint(ss, pkg, dep);
 		} else {
 			/* FIXME: could merge versioning bits too */
-			foreach_array_item(dep, pkg->depends) {
-				name0 = dep->name;
+			apk_array_foreach(dep, pkg->depends) {
+				struct apk_name *name0 = dep->name;
 				if (merge_index_complete(&name0->ss.merge_depends, num_options) &&
 				    name0->ss.requirers == 0) {
 					/* common dependency name with all */
@@ -491,7 +503,7 @@ static void reconsider_name(struct apk_solver_state *ss, struct apk_name *name)
 						   name->name, name0->name);
 					name0->ss.requirers++;
 					name_requirers_changed(ss, name0);
-					foreach_array_item(p, name0->providers)
+					apk_array_foreach(p, name0->providers)
 						inherit_pinning_and_flags(ss, p->pkg, pkg);
 				}
 			}
@@ -500,23 +512,23 @@ static void reconsider_name(struct apk_solver_state *ss, struct apk_name *name)
 		/* provides transitivity */
 		if (merge_index_complete(&pkg->name->ss.merge_provides, num_options))
 			exclude_non_providers(ss, pkg->name, name, pkg->name->ss.has_virtual_provides);
-		foreach_array_item(dep, pkg->provides)
+		apk_array_foreach(dep, pkg->provides)
 			if (merge_index_complete(&dep->name->ss.merge_provides, num_options))
 				exclude_non_providers(ss, dep->name, name, dep->name->ss.has_virtual_provides);
 
 		pkg->name->ss.has_virtual_provides = 0;
-		foreach_array_item(dep, pkg->provides)
+		apk_array_foreach(dep, pkg->provides)
 			dep->name->ss.has_virtual_provides = 0;
 	}
 
 	name->ss.reverse_deps_done = 1;
-	foreach_array_item(pname0, name->rdepends) {
-		name0 = *pname0;
+	apk_array_foreach_item(name0, name->rdepends) {
 		if (name0->ss.seen && !name0->ss.locked) {
 			name->ss.reverse_deps_done = 0;
 			break;
 		}
 	}
+	queue_unresolved(ss, name, reevaluate);
 
 	dbg_printf("reconsider_name: %s [finished], has_options=%d, has_autoselectable=%d, reverse_deps_done=%d\n",
 		name->name, name->ss.has_options, name->ss.has_auto_selectable, name->ss.reverse_deps_done);
@@ -537,15 +549,6 @@ static int compare_providers(struct apk_solver_state *ss,
 	}
 	solver_flags = pkgA->ss.solver_flags | pkgB->ss.solver_flags;
 
-	/* Honor removal preference */
-	if (solver_flags & APK_SOLVERF_REMOVE) {
-		r = (int)(pkgB->ss.solver_flags&APK_SOLVERF_REMOVE) - (int)(pkgA->ss.solver_flags&APK_SOLVERF_REMOVE);
-		if (r) {
-			dbg_printf("    prefer removal hint\n");
-			return r;
-		}
-	}
-
 	/* Latest version required? */
 	if ((solver_flags & APK_SOLVERF_LATEST) &&
 	    (pkgA->ss.pinning_allowed == APK_DEFAULT_PINNING_MASK) &&
@@ -592,7 +595,7 @@ static int compare_providers(struct apk_solver_state *ss,
 		}
 
 		/* Prefer installed on self-upgrade */
-		if ((db->performing_self_upgrade && !(solver_flags & APK_SOLVERF_UPGRADE)) ||
+		if ((db->performing_preupgrade && !(solver_flags & APK_SOLVERF_UPGRADE)) ||
 		    (solver_flags & APK_SOLVERF_INSTALLED)) {
 			r = (pkgA->ipkg != NULL) - (pkgB->ipkg != NULL);
 			if (r) {
@@ -625,7 +628,7 @@ static int compare_providers(struct apk_solver_state *ss,
 		}
 
 		/* Prefer installed */
-		if (!(solver_flags & APK_SOLVERF_UPGRADE)) {
+		if (!(solver_flags & (APK_SOLVERF_REMOVE|APK_SOLVERF_UPGRADE))) {
 			r = (pkgA->ipkg != NULL) - (pkgB->ipkg != NULL);
 			if (r) {
 				dbg_printf("    prefer installed\n");
@@ -635,7 +638,7 @@ static int compare_providers(struct apk_solver_state *ss,
 	}
 
 	/* Select latest by requested name */
-	switch (apk_version_compare_blob(*pA->version, *pB->version)) {
+	switch (apk_version_compare(*pA->version, *pB->version)) {
 	case APK_VERSION_LESS:
 		dbg_printf("    select latest by requested name (less)\n");
 		return -1;
@@ -646,7 +649,7 @@ static int compare_providers(struct apk_solver_state *ss,
 
 	/* Select latest by principal name */
 	if (pkgA->name == pkgB->name) {
-		switch (apk_version_compare_blob(*pkgA->version, *pkgB->version)) {
+		switch (apk_version_compare(*pkgA->version, *pkgB->version)) {
 		case APK_VERSION_LESS:
 			dbg_printf("    select latest by principal name (less)\n");
 			return -1;
@@ -656,13 +659,6 @@ static int compare_providers(struct apk_solver_state *ss,
 		}
 	}
 
-	/* Prefer installed (matches here if upgrading) */
-	r = (pkgA->ipkg != NULL) - (pkgB->ipkg != NULL);
-	if (r) {
-		dbg_printf("    prefer installed (upgrading)\n");
-		return r;
-	}
-
 	/* Prefer highest declared provider priority. */
 	r = pkgA->provider_priority - pkgB->provider_priority;
 	if (r) {
@@ -670,6 +666,15 @@ static int compare_providers(struct apk_solver_state *ss,
 		return r;
 	}
 
+	/* Prefer installed (matches here if upgrading) */
+	if (!(solver_flags & APK_SOLVERF_REMOVE)) {
+		r = (pkgA->ipkg != NULL) - (pkgB->ipkg != NULL);
+		if (r) {
+			dbg_printf("    prefer installed (upgrading)\n");
+			return r;
+		}
+	}
+
 	/* Prefer without errors (mostly if --latest used, and different provider) */
 	r = (int)pkgA->ss.pkg_selectable - (int)pkgB->ss.pkg_selectable;
 	if (r) {
@@ -684,8 +689,6 @@ static int compare_providers(struct apk_solver_state *ss,
 
 static void assign_name(struct apk_solver_state *ss, struct apk_name *name, struct apk_provider p)
 {
-	struct apk_provider *p0;
-
 	if (name->ss.locked) {
 		/* If both are providing this name without version, it's ok */
 		if (p.version == &apk_atom_null &&
@@ -699,8 +702,8 @@ static void assign_name(struct apk_solver_state *ss, struct apk_name *name, stru
 		return;
 	}
 
-	if (p.pkg)
-		dbg_printf("assign %s to "PKG_VER_FMT"\n", name->name, PKG_VER_PRINTF(p.pkg));
+	if (p.pkg) dbg_printf("assign %s to "PKG_VER_FMT"\n", name->name, PKG_VER_PRINTF(p.pkg));
+	else dbg_printf("assign %s to <none>\n", name->name);
 
 	name->ss.locked = 1;
 	name->ss.chosen = p;
@@ -709,11 +712,15 @@ static void assign_name(struct apk_solver_state *ss, struct apk_name *name, stru
 	if (list_hashed(&name->ss.dirty_list))
 		list_del(&name->ss.dirty_list);
 
+	if (p.pkg && !name->ss.requirers && p.pkg->ss.iif_triggered) {
+		apk_array_foreach(dep, p.pkg->install_if)
+			if (!dep->name->ss.locked) apply_constraint(ss, p.pkg, dep);
+	}
+
 	/* disqualify all conflicting packages */
 	if (!ss->ignore_conflict) {
-		foreach_array_item(p0, name->providers) {
-			if (p0->pkg == p.pkg)
-				continue;
+		apk_array_foreach(p0, name->providers) {
+			if (p0->pkg == p.pkg) continue;
 			if (p.version == &apk_atom_null &&
 			    p0->version == &apk_atom_null)
 				continue;
@@ -729,14 +736,14 @@ static void assign_name(struct apk_solver_state *ss, struct apk_name *name, stru
 
 static void select_package(struct apk_solver_state *ss, struct apk_name *name)
 {
-	struct apk_provider chosen = { NULL, &apk_atom_null }, *p;
+	struct apk_provider chosen = { NULL, &apk_atom_null };
 	struct apk_package *pkg = NULL;
-	struct apk_dependency *d;
 
-	dbg_printf("select_package: %s (requirers=%d, autosel=%d, iif=%d)\n", name->name, name->ss.requirers, name->ss.has_auto_selectable, name->ss.has_iif);
+	dbg_printf("select_package: %s (requirers=%d, autosel=%d, iif=%d, order_id=%#x)\n",
+		name->name, name->ss.requirers, name->ss.has_auto_selectable, name->ss.has_iif, name->ss.order_id);
 
 	if (name->ss.requirers || name->ss.has_iif) {
-		foreach_array_item(p, name->providers) {
+		apk_array_foreach(p, name->providers) {
 			dbg_printf("  consider "PKG_VER_FMT" iif_triggered=%d, tag_ok=%d, selectable=%d, available=%d, flags=0x%x, provider_priority=%d, installed=%d\n",
 				PKG_VER_PRINTF(p->pkg),
 				p->pkg->ss.iif_triggered, p->pkg->ss.tag_ok,
@@ -771,10 +778,10 @@ static void select_package(struct apk_solver_state *ss, struct apk_name *name)
 		dbg_printf("selecting: " PKG_VER_FMT ", available: %d\n", PKG_VER_PRINTF(pkg), pkg->ss.pkg_selectable);
 
 		assign_name(ss, pkg->name, APK_PROVIDER_FROM_PACKAGE(pkg));
-		foreach_array_item(d, pkg->provides)
+		apk_array_foreach(d, pkg->provides)
 			assign_name(ss, d->name, APK_PROVIDER_FROM_PROVIDES(pkg, d));
 
-		foreach_array_item(d, pkg->depends)
+		apk_array_foreach(d, pkg->depends)
 			apply_constraint(ss, pkg, d);
 	} else {
 		dbg_printf("selecting: %s [unassigned]\n", name->name);
@@ -791,14 +798,13 @@ static void record_change(struct apk_solver_state *ss, struct apk_package *opkg,
 	struct apk_changeset *changeset = ss->changeset;
 	struct apk_change *change;
 
-	change = apk_change_array_add(&changeset->changes);
-	*change = (struct apk_change) {
+	change = apk_change_array_add(&changeset->changes, (struct apk_change) {
 		.old_pkg = opkg,
 		.old_repository_tag = opkg ? opkg->ipkg->repository_tag : 0,
 		.new_pkg = npkg,
 		.new_repository_tag = npkg ? get_tag(ss->db, npkg->ss.pinning_allowed, get_pkg_repos(ss->db, npkg)) : 0,
 		.reinstall = npkg ? !!(npkg->ss.solver_flags & APK_SOLVERF_REINSTALL) : 0,
-	};
+	});
 	if (npkg == NULL)
 		changeset->num_remove++;
 	else if (opkg == NULL)
@@ -813,22 +819,18 @@ static void cset_gen_dep(struct apk_solver_state *ss, struct apk_package *ppkg,
 
 static void cset_track_deps_added(struct apk_package *pkg)
 {
-	struct apk_dependency *d;
-
-	foreach_array_item(d, pkg->depends) {
-		if (d->conflict || !d->name->ss.installed_name)
-			continue;
+	apk_array_foreach(d, pkg->depends) {
+		if (apk_dep_conflict(d) || !d->name->ss.installed_name) continue;
 		d->name->ss.installed_name->ss.requirers++;
 	}
 }
 
 static void cset_track_deps_removed(struct apk_solver_state *ss, struct apk_package *pkg)
 {
-	struct apk_dependency *d;
 	struct apk_package *pkg0;
 
-	foreach_array_item(d, pkg->depends) {
-		if (d->conflict || !d->name->ss.installed_name)
+	apk_array_foreach(d, pkg->depends) {
+		if (apk_dep_conflict(d) || !d->name->ss.installed_name)
 			continue;
 		if (--d->name->ss.installed_name->ss.requirers > 0)
 			continue;
@@ -852,15 +854,13 @@ static void cset_check_removal_by_deps(struct apk_solver_state *ss, struct apk_p
 static void cset_check_install_by_iif(struct apk_solver_state *ss, struct apk_name *name)
 {
 	struct apk_package *pkg = name->ss.chosen.pkg;
-	struct apk_dependency *dep0;
 
-	if (pkg == NULL || !name->ss.seen || name->ss.in_changeset)
-		return;
+	if (!pkg || !name->ss.seen || name->ss.changeset_processed) return;
 
-	foreach_array_item(dep0, pkg->install_if) {
+	apk_array_foreach(dep0, pkg->install_if) {
 		struct apk_name *name0 = dep0->name;
-		if (!name0->ss.in_changeset) return;
-		if (!apk_dep_is_provided(dep0, &name0->ss.chosen)) return;
+		if (!apk_dep_conflict(dep0) && !name0->ss.changeset_processed) return;
+		if (!apk_dep_is_provided(pkg, dep0, &name0->ss.chosen)) return;
 	}
 	cset_gen_name_change(ss, name);
 }
@@ -868,14 +868,13 @@ static void cset_check_install_by_iif(struct apk_solver_state *ss, struct apk_na
 static void cset_check_removal_by_iif(struct apk_solver_state *ss, struct apk_name *name)
 {
 	struct apk_package *pkg = name->ss.installed_pkg;
-	struct apk_dependency *dep0;
 
-	if (pkg == NULL || name->ss.in_changeset || name->ss.chosen.pkg != NULL)
-		return;
+	if (!pkg || name->ss.chosen.pkg) return;
+	if (name->ss.changeset_processed || name->ss.changeset_removed) return;
 
-	foreach_array_item(dep0, pkg->install_if) {
-		if (dep0->name->ss.in_changeset &&
-		    dep0->name->ss.chosen.pkg == NULL) {
+	apk_array_foreach(dep0, pkg->install_if) {
+		struct apk_name *name0 = dep0->name;
+		if (name0->ss.changeset_removed && !name0->ss.chosen.pkg) {
 			cset_check_removal_by_deps(ss, pkg);
 			return;
 		}
@@ -884,23 +883,16 @@ static void cset_check_removal_by_iif(struct apk_solver_state *ss, struct apk_na
 
 static void cset_check_by_reverse_iif(struct apk_solver_state *ss, struct apk_package *pkg, void (*cb)(struct apk_solver_state *ss, struct apk_name *))
 {
-	struct apk_name **pname;
-	struct apk_dependency *d;
-
 	if (!pkg) return;
-	foreach_array_item(pname, pkg->name->rinstall_if)
-		cb(ss, *pname);
-	foreach_array_item(d, pkg->provides)
-		foreach_array_item(pname, d->name->rinstall_if)
-			cb(ss, *pname);
+	apk_array_foreach_item(name, pkg->name->rinstall_if) cb(ss, name);
+	apk_array_foreach(d, pkg->provides)
+		apk_array_foreach_item(name, d->name->rinstall_if) cb(ss, name);
 }
 
-static void cset_gen_name_remove_orphan(struct apk_solver_state *ss, struct apk_name *name)
+static void cset_gen_name_preprocess(struct apk_solver_state *ss, struct apk_name *name)
 {
-	struct apk_provider *p;
-
-	if (name->ss.in_changeset) return;
-	name->ss.in_changeset = 1;
+	if (name->ss.changeset_processed) return;
+	name->ss.changeset_processed = 1;
 
 	dbg_printf("cset_gen_name_remove_orphans: %s\n", name->name);
 
@@ -910,7 +902,7 @@ static void cset_gen_name_remove_orphan(struct apk_solver_state *ss, struct apk_
 		cset_gen_name_remove(ss, name->ss.installed_pkg);
 
 	/* Remove any package that provides this name and is due to be deleted */
-	foreach_array_item(p, name->providers) {
+	apk_array_foreach(p, name->providers) {
 		struct apk_package *pkg0 = p->pkg;
 		struct apk_name *name0 = pkg0->name;
 		if (name0->ss.installed_pkg == pkg0 && name0->ss.chosen.pkg == NULL)
@@ -921,25 +913,24 @@ static void cset_gen_name_remove_orphan(struct apk_solver_state *ss, struct apk_
 static void cset_gen_name_change(struct apk_solver_state *ss, struct apk_name *name)
 {
 	struct apk_package *pkg, *opkg;
-	struct apk_dependency *d;
 
-	if (name->ss.in_changeset) return;
+	if (name->ss.changeset_processed) return;
 
 	dbg_printf("cset_gen: processing: %s\n", name->name);
-	cset_gen_name_remove_orphan(ss, name);
+	cset_gen_name_preprocess(ss, name);
 
 	pkg = name->ss.chosen.pkg;
 	if (!pkg || pkg->ss.in_changeset) return;
-	pkg->ss.in_changeset = 1;
 
-	cset_gen_name_remove_orphan(ss, pkg->name);
-	foreach_array_item(d, pkg->provides)
-		cset_gen_name_remove_orphan(ss, d->name);
+	pkg->ss.in_changeset = 1;
+	cset_gen_name_preprocess(ss, pkg->name);
+	apk_array_foreach(d, pkg->provides)
+		cset_gen_name_preprocess(ss, d->name);
 
 	opkg = pkg->name->ss.installed_pkg;
 	cset_check_by_reverse_iif(ss, opkg, cset_check_removal_by_iif);
 
-	foreach_array_item(d, pkg->depends)
+	apk_array_foreach(d, pkg->depends)
 		cset_gen_dep(ss, pkg, d);
 
 	dbg_printf("cset_gen: selecting: "PKG_VER_FMT"%s\n", PKG_VER_PRINTF(pkg), pkg->ss.pkg_selectable ? "" : " [NOT SELECTABLE]");
@@ -966,7 +957,7 @@ static void cset_gen_name_remove(struct apk_solver_state *ss, struct apk_package
 	     name->ss.chosen.pkg->name == name))
 		return;
 
-	name->ss.in_changeset = 1;
+	name->ss.changeset_removed = 1;
 	pkg->ss.in_changeset = 1;
 	apk_pkg_foreach_reverse_dependency(pkg, APK_FOREACH_INSTALLED|APK_DEP_SATISFIES, cset_gen_name_remove0, ss);
 	cset_check_by_reverse_iif(ss, pkg, cset_check_removal_by_iif);
@@ -980,10 +971,10 @@ static void cset_gen_dep(struct apk_solver_state *ss, struct apk_package *ppkg,
 	struct apk_name *name = dep->name;
 	struct apk_package *pkg = name->ss.chosen.pkg;
 
-	if (dep->conflict && ss->ignore_conflict)
+	if (apk_dep_conflict(dep) && ss->ignore_conflict)
 		return;
 
-	if (!apk_dep_is_provided(dep, &name->ss.chosen))
+	if (!apk_dep_is_provided(ppkg, dep, &name->ss.chosen))
 		mark_error(ss, ppkg, "unfulfilled dependency");
 
 	cset_gen_name_change(ss, name);
@@ -1006,16 +997,15 @@ static void generate_changeset(struct apk_solver_state *ss, struct apk_dependenc
 	struct apk_changeset *changeset = ss->changeset;
 	struct apk_package *pkg;
 	struct apk_installed_package *ipkg;
-	struct apk_dependency *d;
 
-	apk_change_array_init(&changeset->changes);
+	apk_array_truncate(changeset->changes, 0);
 
 	apk_hash_foreach(&ss->db->available.names, cset_reset_name, NULL);
 	list_for_each_entry(ipkg, &ss->db->installed.packages, installed_pkgs_list) {
 		pkg = ipkg->pkg;
 		pkg->name->ss.installed_pkg = pkg;
 		pkg->name->ss.installed_name = pkg->name;
-		foreach_array_item(d, pkg->provides)
+		apk_array_foreach(d, pkg->provides)
 			if (d->version != &apk_atom_null)
 				d->name->ss.installed_name = pkg->name;
 	}
@@ -1024,7 +1014,7 @@ static void generate_changeset(struct apk_solver_state *ss, struct apk_dependenc
 	list_for_each_entry(ipkg, &ss->db->installed.packages, installed_pkgs_list)
 		cset_check_removal_by_deps(ss, ipkg->pkg);
 
-	foreach_array_item(d, world)
+	apk_array_foreach(d, world)
 		cset_gen_dep(ss, NULL, d);
 
 	/* NOTE: We used to call cset_gen_name_remove() directly here.  While slightly faster, this clobbered
@@ -1060,15 +1050,24 @@ static int cmp_pkgname(const void *p1, const void *p2)
 	return apk_name_cmp_display(d1->name, d2->name);
 }
 
-static int compare_name_dequeue(const struct apk_name *a, const struct apk_name *b)
+static struct apk_name *dequeue_next_name(struct apk_solver_state *ss)
 {
-	int r = (int)b->ss.has_auto_selectable - (int)a->ss.has_auto_selectable;
-	if (r) return r;
-
-	r = !!a->solver_flags_set - !!b->solver_flags_set;
-	if (r) return -r;
-
-	return b->ss.order_id - a->ss.order_id;
+	if (!list_empty(&ss->resolvenow_head)) {
+		struct apk_name *name = list_pop(&ss->resolvenow_head, struct apk_name, ss.unresolved_list);
+		dbg_printf("name <%s> selected from resolvenow list\n", name->name);
+		return name;
+	}
+	if (!list_empty(&ss->selectable_head)) {
+		struct apk_name *name = list_pop(&ss->selectable_head, struct apk_name, ss.unresolved_list);
+		dbg_printf("name <%s> selected from selectable list\n", name->name);
+		return name;
+	}
+	if (!list_empty(&ss->unresolved_head)) {
+		struct apk_name *name = list_pop(&ss->unresolved_head, struct apk_name, ss.unresolved_list);
+		dbg_printf("name <%s> selected from unresolved list\n", name->name);
+		return name;
+	}
+	return NULL;
 }
 
 int apk_solver_solve(struct apk_database *db,
@@ -1076,12 +1075,11 @@ int apk_solver_solve(struct apk_database *db,
 		     struct apk_dependency_array *world,
 		     struct apk_changeset *changeset)
 {
-	struct apk_name *name, *name0;
+	struct apk_name *name;
 	struct apk_package *pkg;
 	struct apk_solver_state ss_data, *ss = &ss_data;
-	struct apk_dependency *d;
 
-	qsort(world->item, world->num, sizeof(world->item[0]), cmp_pkgname);
+	apk_array_qsort(world, cmp_pkgname);
 
 restart:
 	memset(ss, 0, sizeof(*ss));
@@ -1091,15 +1089,17 @@ restart:
 	ss->ignore_conflict = !!(solver_flags & APK_SOLVERF_IGNORE_CONFLICT);
 	list_init(&ss->dirty_head);
 	list_init(&ss->unresolved_head);
+	list_init(&ss->selectable_head);
+	list_init(&ss->resolvenow_head);
 
 	dbg_printf("discovering world\n");
 	ss->solver_flags_inherit = solver_flags;
-	foreach_array_item(d, world) {
+	apk_array_foreach(d, world) {
 		if (!d->broken)
 			discover_name(ss, d->name);
 	}
 	dbg_printf("applying world\n");
-	foreach_array_item(d, world) {
+	apk_array_foreach(d, world) {
 		if (!d->broken) {
 			ss->pinning_inherit = BIT(d->repository_tag);
 			apply_constraint(ss, NULL, d);
@@ -1114,28 +1114,16 @@ restart:
 			name = list_pop(&ss->dirty_head, struct apk_name, ss.dirty_list);
 			reconsider_name(ss, name);
 		}
-
-		name = NULL;
-		list_for_each_entry(name0, &ss->unresolved_head, ss.unresolved_list) {
-			if (name0->ss.reverse_deps_done && name0->ss.requirers &&
-			    name0->ss.has_auto_selectable && !name0->ss.has_options) {
-				name = name0;
-				dbg_printf("name <%s> fast selected\n", name->name);
-				break;
-			}
-			if (!name || compare_name_dequeue(name0, name) < 0)
-				name = name0;
-		}
+		name = dequeue_next_name(ss);
 		if (name == NULL)
 			break;
-
 		select_package(ss, name);
 	} while (1);
 
 	generate_changeset(ss, world);
 
-	if (ss->errors && (apk_force & APK_FORCE_BROKEN_WORLD)) {
-		foreach_array_item(d, world) {
+	if (ss->errors && (db->ctx->force & APK_FORCE_BROKEN_WORLD)) {
+		apk_array_foreach(d, world) {
 			name = d->name;
 			pkg = name->ss.chosen.pkg;
 			if (pkg == NULL || pkg->ss.error) {
@@ -1148,6 +1136,11 @@ restart:
 		goto restart;
 	}
 
+	apk_array_foreach(d, world) {
+		if (!d->name->ss.chosen.pkg) continue;
+		d->layer = d->name->ss.chosen.pkg->layer;
+	}
+
 	apk_hash_foreach(&db->available.names, free_name, NULL);
 	apk_hash_foreach(&db->available.packages, free_package, NULL);
 	dbg_printf("solver done, errors=%d\n", ss->errors);
